// (C) 2005 Dmytry Lavrov.
// You are free to modify and redistribute this program code,
// but this comment block must remain unchanged and be present in any redistribution.

// Possible improvements: maybe use vector and binary search instead of std::set ?
// Maybe use "smarter-than-human" context-sensitive algorithms?

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <algorithm>
#include <map>
#include <set>
#include <cctype>

struct Word {
    std::string key;
    std::string original_word;
    int use_count;
    Word() {
        use_count=0;
    }
    Word(const std::string &s) {
        use_count=1;
        original_word=s;
        key=s;
        if(key.size()>3) {
            std::sort(key.begin()+1,key.end()-1);
        };
    }
    bool operator<(const Word &other)const {
        return
            key<other.key || (key==other.key && original_word<other.original_word);
    }
};
class LangDataBase {
    typedef std::set<Word> set_type;
    set_type word_set;
    void Add(Word word) {
        set_type::iterator i=word_set.find(word);
        if(i==word_set.end()) {
            word_set.insert(word_set.end(),word);
        }
        else {
          // increment use count for this word
           *(const_cast<int *>(&((*i).use_count))) += word.use_count;
        }
    }
    void Add(const std::string &s) {
      if(s.length()>3){
        Word word(s);
        Add(word);
      }
    }
public:
    void Save(std::ostream &s) {
      for(set_type::iterator i=word_set.begin();i!=word_set.end();++i){
        s<<
         /* (*i).key<<" "<< /* */
        (*i).original_word<<" "<<(*i).use_count<<std::endl;
      }
      //PrintStats(std::cout);
    }
    void Load(std::istream &s) {
      while(s.good()){
        std::string word_s;
        //getline(s,word_s,' ');
        //if(!s.good())return;
        getline(s,word_s,' ');
        if(!s.good())return;
        int ucount;
        s>>ucount;
        if(word_s.length()>3){
        Word word(word_s);
        word.use_count=ucount;
        Add(word);
        }
        // remove end of line
        if(s.good())getline(s,word_s,'\n');
      }
    }
    void Update(std::istream &s) {
      std::string word;
      while(s.good()){
        char c;
        s.get(c);
        c=tolower(c);
        if(c>='a' && c<='z'){
          word+=c;
        }else
        if(!word.empty()){
          Add(word);
          word.clear();
        };
      }

      if(!word.empty()){
        Add(word);
        word.clear();
      };
    }
    std::string DecipherWord(const std::string &s){
      if(s.size()<=3)return s;
      Word w(s);
      w.original_word="";
      set_type::iterator l=word_set.lower_bound(w);
      w.original_word=(unsigned char)(255);
      set_type::iterator h=word_set.upper_bound(w);
      int max_use_count=0;
      set_type::iterator maxi=word_set.end();
      set_type::iterator i;
      for(i=l;i!=h;++i){
        if((*i).key==w.key){
          // if original word is s then it weren't encrypted in first place
          if((*i).original_word==s){
            maxi=i;
            break;
          }
          // otherwise select one with biggest use_count
          if((*i).use_count>max_use_count){
            max_use_count=(*i).use_count;
            maxi=i;
          }

        }
      }
      if(maxi!=word_set.end()){
        return (*maxi).original_word;
      }else {
        return s;
      }
    }

    void Decipher(std::istream &is,std::ostream &os){
      std::string word;
      while(is.good()){
        char c=255;
        is.get(c);
        if(c==255)break;
        char lc=tolower(c);
        if(lc>='a' && lc<='z'){
          word+=lc;
        }else{
         if(!word.empty()){
          os<<DecipherWord(word);
          word.clear();
         }
         os<<c;
        }
      }
      if(!word.empty()){
          os<<DecipherWord(word);
      }
    }
    void PrintStats(std::ostream &os){
      int collisions=0;
      Word prev_word;
      for(set_type::iterator i=word_set.begin();i!=word_set.end();++i){
        if((*i).key==prev_word.key){
          os<<"Collision: "<<prev_word.original_word<<"("<<prev_word.use_count<<")"<<" with "<<(*i).original_word<<"("<<(*i).use_count<<")"<<std::endl;
          ++collisions;
        }
        prev_word= *i;
      }
      os<<"Words loaded(excluding 3 letter words and shorter)="<<word_set.size()<<std::endl;
      os<<"Collisions="<<collisions<<std::endl;
    }
};

LangDataBase db;
enum{mode_undefined,mode_load, mode_save, mode_add, mode_decipher};

int main(int argc, char **argv) {

    if(argc==1){
      show_usage:
      std::cout << "RaedTaht v 1.0 , (C) 2005 Dmytry Lavrov." << std::endl;
      std::cout << "Parameters:" << std::endl;
      std::cout << "raedtaht {-l files | -s files | -a files | -d files | -i} ..." << std::endl;
      std::cout << " -l = load dictionary file \n"
      " -s = save dictionary file \n"
      " -a = load some arbitrary text and add words to dictionary \n"
      " -d = decipher files (output to console) \n"
      " -i = print some information such as number of 'collisions' \n"
      " Commands are executed in the specified order.(so to decipher file you must load dictionary, then decipher file with -d) \n"
      << std::endl;
      std::cout << "Usage Examples:" << std::endl;
      std::cout << "raedtaht -a book1.txt book2.txt book3.txt -l dictionary.txt -s dictionary.txt" << std::endl;
      std::cout << "will merge words from book1.txt,book2.txt,book3.txt with dictionary.txt" << std::endl<< std::endl;
      std::cout << "raedtaht -l dictionary -d sutdy.txt" << std::endl;
      std::cout << "will load dictionary and then decipher sutdy.txt" << std::endl;
      return 1;
    }
    int mode=mode_undefined;
    for(int i=1;i<argc;++i){
      std::string s=argv[i];
      if(s=="-a"){
        mode=mode_add;
      }else
      if(s=="-l"){
        mode=mode_load;
      }else
      if(s=="-s"){
        mode=mode_save;
      }else
      if(s=="-d"){
        mode=mode_decipher;
      }else
      if(s=="-i"){
        //mode=mode_decipher;
        db.PrintStats(std::cout);
      }else
      {
        switch(mode){
          case mode_undefined:goto show_usage;
          case mode_add:{
            std::ifstream is(s.c_str());
            db.Update(is);
            break;
          }
          case mode_load:{
            std::ifstream is(s.c_str());
            db.Load(is);
            break;
          }
          case mode_save:{
            std::ofstream os(s.c_str());
            db.Save(os);
            break;
          }
          case mode_decipher:{
            //std::cout<<"decipher is not implemented yet"<<std::endl;
            std::ifstream is(s.c_str());
            db.Decipher(is,std::cout);
          }
        }
      }
    }

    return 0;
}
